#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include "basic.h"

static int verbose = 0 ;

double norm(double *t, int sz) {
  double r = 0.0 ;
  for (int k = 0 ; k < sz ; k++) {
    double x = t[k] ;
    r += x * x;
  }
  return r ;
}

double norm_partial(double *t, int kmin, int kmax) {
  double r = 0.0 ;
  for (int k = kmin ; k < kmax ; k++) {
    double x = t[k] ;
    r += x * x ;
  }
  return r ;
}

typedef struct {
  int nprocs, sz ;
  pthread_barrier_t *b ;
  double *r, *t ;
  double result ;
} common_t ;

typedef struct {
  int id ; 
  common_t *common ; 
} ctx_t ;

void *f(void *_p) {
  ctx_t *p = _p ;
  common_t *q = p->common ;

  // Task 1
  int chunk = q->sz / q->nprocs ;
  int kmin = p->id * chunk ;
  int kmax = kmin+chunk ;
  if (p->id == q->nprocs-1) kmax = q->sz ;
  q->r[p->id] = norm_partial(q->t,kmin,kmax) ;
  (void)wait_barrier(q->b) ;

  // Task 2
  if (p->id == 0) {
    q->result = 0.0 ;
    for (int k = 0 ; k < q->nprocs ; k++) q->result += q->r[k] ;
  }
  return NULL ;
}


void run(int sz, int nprocs) {
  pthread_barrier_t *b = alloc_barrier(nprocs) ;
  double *t = malloc_check(sizeof(t[0])*sz) ;
  double *r = malloc_check(sizeof(r[0])*nprocs) ;
  for (int k = 0 ; k < sz ; k++) t[k] = k ;

  common_t c ;
  c.nprocs = nprocs, c.sz = sz ;
  c.b = b  ; c.t = t ; c.r = r ;
  
  tsc_t t0 = timeofday() ;
  pthread_t th[nprocs] ;
  ctx_t a[nprocs] ;
  for (int k = 0 ; k < nprocs ; k++) {
    ctx_t *p = &a[k] ;
    p->id = k ;
    p->common = &c ;
    create_thread(&th[k],f,p) ;
  }
  for (int k = 0 ; k < nprocs ; k++)join_thread(&th[k]) ;
  tsc_t t1 = timeofday() ;
  printf("Parallel (nprocs=%i) %g [dt=%g]\n",nprocs,c.result,tsc_millions(t1-t0)) ;
  t0 = timeofday() ;
  double seq_result = norm(t,sz) ;
  t1 = timeofday() ;
  printf("Sequential %g [dt=%g]\n",seq_result,tsc_millions(t1-t0))  ;
  free(r) ; free(t) ; free_barrier(b) ;
}



void usage(char *pgm) {
  fprintf(stderr,"Usage: %s [-v]* [sz] [nprocs]\n",pgm) ;
}

int main(int argc, char **argv) {
  char * pgm = argv[0] ;
  int sz = 1000000 ;
  int nprocs = 2 ;

  ++argv ;
  --argc ;
  while (argc > 0) {
    char *arg = *argv ;
    if (*arg == '-') {
      arg++ ;
      if (*arg == 'v') verbose++ ;
      else usage(pgm) ;
    } else {
      break ;
    }
    ++argv;
    --argc ;
  }
  if (argc == 1) {    
    sz = atoi(argv[0]) ;
  } else if (argc > 1) {
    sz = atoi(argv[0]) ;
    nprocs = atoi(argv[1]) ;
  }
  run(sz,nprocs) ;
  return 0 ;
}

