#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <inttypes.h>
#include <errno.h>
#include "basic.h"
#include "treiber_hazard.h"

#define CAS(x,y,z) __sync_bool_compare_and_swap(x,y,z)

static int dbg = 0 ;

/* List of nodes, with count */
#define MAGIC ((void *)0x12345)

static const free_list_t empty_free =  { NULL, 0, } ;

static void push_free(free_list_t *f,stack_node_t *p) {
  p->next = f->p ;
  f->p = p ;
  f->count++ ;  
}


/* Local free lists */
__thread free_list_t my_free_list = { NULL, 0, } ;

static const int NCHUNK = 1 ;

static stack_node_t *alloc_stack_node(void *val) {
  if (my_free_list.count == 0) {
    for (int k = 0 ; k < NCHUNK ; k++) {
      stack_node_t *q = malloc_check(sizeof(stack_node_t)) ;
      q->val = MAGIC ;
      push_free(&my_free_list,q) ;
    }
  }
  stack_node_t *r = my_free_list.p ;
  my_free_list.p = r->next ;
  my_free_list.count-- ;
  r->val = val ; r->next = NULL ;
  return r ;
}

void flush_free_list(void) {
  stack_node_t *p = my_free_list.p ;
  while (p) {
    stack_node_t *q = p->next ;
    if (p->val != MAGIC) {
      fprintf(stderr,"NO MAGIC %p\n",p) ;
    }
    free(p) ;
    p = q ;
  }
  my_free_list.p = NULL ;
  my_free_list.count = 0 ;
}

static void free_stack_node(stack_node_t *p) {
  if (p->val == MAGIC) {
    fprintf(stderr, "DOUBLE FREE: %p\n",p) ;
  }
  p->val = MAGIC ;
  push_free(&my_free_list,p) ;
  if (my_free_list.count >= NCHUNK) flush_free_list() ;
}

/* Nodes */
#define SORTED 1

#ifdef SORTED
static int cmp_ptr(const void *_p, const void *_q) {
  stack_node_t *p = *(stack_node_t **)_p ;
  stack_node_t *q = *(stack_node_t **)_q ;
  return (p > q) - (p < q) ; /* compare pointers... */
}
#endif

static int mem(stack_node_t *t[], int sz, stack_node_t *p) {
#ifdef SORTED
  for (int k = 0 ; k < sz-1 ; k++) {
    if (t[k] > t[k+1]) { 
      fprintf(stderr,"NOT SORTED\n") ; exit(1) ;
    }
  }
  int a = 0 ;
  int b = sz ;
  while (a < b) {
    int m = (a+b)/2 ;
    stack_node_t *q = t[m] ;
    if (p < q) b = m ;
    else if (p > q) a = m+1 ;
    else return 1 ;
  }
#else
  for (int k = 0 ; k < sz ; k++)
    if (p == t[k]) return 1 ;
#endif
  return 0 ;
}

static void scan_hazard(stack_t *s,int id) {
/* First phase, collect hazard pointers */
  stack_node_t *t[s->nprocs] ;
  int sz = 0 ;
  for (int k = 0 ; k < s->nprocs ; k++) {
    stack_node_t *p = s->hazard[k] ;
    if (p != NULL) t[sz++] = p ;
  }
#ifdef SORTED
  qsort(t,sz,sizeof(t[0]),cmp_ptr) ;
#endif
/* Second phase, free non-hazard, keep hazard */  
  free_list_t f = empty_free ;
  for (stack_node_t *p = s->retired[id].p ; p ; p = p->next) {
    if (mem(t,sz,p)) {
      if (dbg) {
        fprintf(stderr,"%02i KEEP %p\n",id,p) ;
      }
      push_free(&f,p) ;
    } else {
      free_stack_node(p) ;
    }
  }
  s->retired[id] = f ;
}

static void retire_stack_node(stack_t *s,int id,stack_node_t *p) {
  push_free(&s->retired[id],p) ;
  if (s->retired[id].count > s->nprocs) scan_hazard(s,id) ;
}

stack_t *alloc_stack(int nprocs) {
  stack_t *r =  malloc_check(sizeof(*r)) ;
  r->top = NULL ;
  r->nprocs = nprocs ;
  r->hazard = calloc_check(nprocs,sizeof(r->hazard[0])) ;
  r->retired = calloc_check(nprocs,sizeof(r->retired[0])) ;
  r->mutex = alloc_mutex() ;
  return r ;
}

void free_stack(stack_t *p) {
  free(p->hazard) ;
  for (int k = 0 ; k < p->nprocs ; k++) {
    stack_node_t *q = p->retired[k].p ;
    if (dbg) {
      int c = p->retired[k].count ;
      if (c > 0) {
        fprintf(stderr,"%02i FREE COUNT %i\n",k,c) ;
      }
    }
    while (q) {
      stack_node_t *r = q->next ;
      free_stack_node(q) ;
      q = r ;
    }
  }
  flush_free_list() ;
  free(p->retired) ;
  free_mutex(p->mutex) ;
  free(p) ;
}

void *pop(stack_t *p, int id) {
  for ( ; ; ) {
    SYNC() ;
    stack_node_t *old = p->top ; // Old top node.
    if (old == NULL) return NULL ;
    SYNC() ;
    p->hazard[id] = old ;
    SYNC() ;
    if (p->top != old) continue ;
    SYNC() ; 
    stack_node_t *next = old->next ;
    SYNC() ;
    if (CAS(&p->top,old,next)) { // Do pop
      SYNC() ;
      if (dbg > 1) fprintf(stderr,"POP: %p\n",old) ;
      void *r = old->val ;         // Clean up and return
      retire_stack_node(p,id,old) ;
      return r ;
    }
  }
}


void push(stack_t *p, void *v) {
  stack_node_t *q = alloc_stack_node(v), *old ;
  do {
    SYNC() ;
    old = p->top ;
    SYNC() ;
    q->next = old ;                        
    SYNC() ;
  } while (!CAS(&p->top,old,q)) ; // Do push
  SYNC() ;
  if (dbg > 1) fprintf(stderr,"PUSH: %p\n",q) ;
}
