#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <inttypes.h>
#include "basic.h"

static int verbose = 0 ;


/* Barrier */

typedef struct {
  unsigned int nprocs ;
  volatile unsigned int count ;
  volatile int sense ;
} barrier_t ; 


barrier_t *alloc_my_barrier(unsigned int nprocs) {
  barrier_t *p = malloc_check(sizeof(*p)) ;
  p->nprocs = p->count = nprocs ;
  p->sense = 0 ;
  return p ;
}

void free_my_barrier(barrier_t *p) {
  free(p) ;
}

/* Coded in C */
int wait_my_barrier(barrier_t *p) {
  int r = 0 ;
  int sense = p->sense ;
  int rem = DECR(&p->count) ;
  SYNC() ;
  if (rem == 0) {
    r = 1 ;
    p->count = p->nprocs ;
    SYNC() ;
    p->sense = !sense ;
  } else {
    unsigned int c = 0 ;
    while (p->sense == sense) { if (++c == 8000000U) { c = 0 ; printf("+") ; fflush(stdout) ; } }
  }
  SYNC() ;
  return r ;
}

#define LINE 128
#define SZ (LINE/sizeof(int))

typedef struct {
  int y,r2;
} log_t ;

static void pp_log(FILE *chan,log_t *p) {
  fprintf(chan,"y=%i, r2=%i\n",p->y,p->r2) ;
}

typedef uint64_t tb_t ;
#define PTB PRIu64


#if 1
inline static tb_t read_timebase(void) {
  unsigned int a,d;
  asm __volatile__ ("rdtsc" : "=a" (a), "=d" (d));
  return ((tb_t)a) | (((tb_t)d)<<32);
}
#else

inline static tb_t read_timebase(void) {
  tb_t r;
  asm __volatile__ ("mftb %[r1]" :[r1] "=r" (r) : : "memory");
  return r;
}

#endif



/* Instance context */
typedef struct {
  void *mem ;
  int *cache ;
  log_t *out ;
  tb_t next_tb;
  barrier_t *b;
} ctx_t ;

ctx_t *alloc_ctx(int nlines,int log_sz) {
  ctx_t *r = malloc_check(sizeof(*r)) ;
  r->mem = malloc_check((nlines+1)*LINE) ;
  intptr_t x = (intptr_t)r->mem ;
  x += LINE-1 ;
  x /= LINE ;
  x *= LINE ;
  r->cache = (int *)x ;
  r->out = malloc_check(log_sz*sizeof(r->out[0])) ;
  return r ;
}

void free_ctx(ctx_t *p) {
  free(p->mem) ;
  free(p->out) ;
  free(p) ;
}

typedef struct {
  int id ;
  int role ;
  barrier_t *b ;
  int size ;
  int *cache ;
  log_t *out ;
  tb_t *next_tb;
} arg_t ;

arg_t *alloc_arg(int size) {
  arg_t *r = malloc_check(sizeof(*r));
  r->size = size ;
  return r ;
}

void free_arg(arg_t*p) { free(p); }

void *thread(void *_a) {
  arg_t *a = (arg_t *)_a ;
  int *x=a->cache,*y=a->cache+1 ;
  log_t *out = a->out ;
  int _delay = 2048 ;
  wait_my_barrier(a->b) ;
  for (int k = 0  ; k < a->size ; k++) {
    if (a->role == 0) {  *x = *y = 0 ; }
    if (k % 2 == a->role) *a->next_tb = read_timebase() ;
    wait_my_barrier(a->b) ;

    tb_t _tb0 = *a->next_tb;
    int _delta;
    do { _delta =  read_timebase() - _tb0; } while (_delta < _delay);    

    switch (a->role) {
    case 0:
      *x = 1 ;
      *y = 1 ;
      break ;
    case 1:
      *y = 2 ;
      int r2 = *x ;
      out[k].r2 = r2 ;
      break;
    }
    wait_my_barrier(a->b) ;
    if (a->role == 0) {
      out[k].y = *y ; 
    }
    //    delay(10) ;
    //    if (a->role == 0) _delay += 128 ;
  }
  return NULL ;
}

void run(int size, int ninst) {
  int count[3][2] ;
  for (int k1 = 0 ; k1 < 3 ; k1++)
    for (int k2 = 0 ; k2 < 2 ; k2++)
      count[k1][k2] = 0 ;
  int nt = 2 ;
  int nth = nt * ninst ;
  pthread_t th[nth] ;
  ctx_t *ctx[ninst] ;
  arg_t *arg[nth] ;
  int id = 0 ;
  for (int i = 0 ; i < ninst ; i++) {
    ctx[i] = alloc_ctx(16,size) ;
    ctx_t *c = ctx[i] ;
    c->b = alloc_my_barrier(nt) ;
    for (int role = 0 ; role < nt ; role++) {
      arg[id] = alloc_arg(size) ;
      arg_t *a = arg[id] ;
      a->id = id ; a->role = role ;
      a->b = c->b ; a->cache = c->cache ; a->out = c->out ;
      a->next_tb = &c->next_tb ;
      id++ ;
    }
  }

  for (int nrun = 1; ; nrun++) {
    for (int n = 0 ; n < nth ; n++)
      create_thread(&th[n],thread,arg[n]) ;

    for (int n = 0 ; n < nth ; n++)
      (void)join_thread(&th[n]) ;

    if (verbose) fprintf(stderr,"+++++++++++++++\n") ;
    for (int k = 0 ; k < size ; k++) {
      for (int i = 0 ; i < ninst ; i++) {
        log_t *p = &ctx[i]->out[k] ;
        if (verbose) pp_log(stderr,p) ;
        count[p->y][p->r2]++ ;
      }
    }
    if (verbose) fflush(stderr) ;

    printf("** %i **\n",nrun) ;
    int nz = 0 ;
    for (int k1 = 1 ; k1 < 3 ; k1++)
      for (int k2 = 0 ; k2 < 2 ; k2++) {
        int c = count[k1][k2] ;
        if (c > 0) nz++ ;
        printf("%-5i> y=%i r2=%i\n",c,k1,k2) ;
      }
    if (nz >= 4) break ;
  }
  for (int i = 0 ; i < ninst ; i++) {
    free_my_barrier(ctx[i]->b) ;
    free_ctx(ctx[i]) ;
  }
  for (int n = 0 ; n < nth ; n++)
    free_arg(arg[n]) ;
}


void usage(char *pgm) {
  fprintf(stderr,"Usage: %s [-v]* [n]\n",pgm) ;
}

int main(int argc, char **argv) {
  char * pgm = argv[0] ;
  int size = 16 ;
  
  ++argv ;
  --argc ;
  while (argc > 0) {
    char *arg = *argv ;
    if (*arg == '-') {
      arg++ ;
      if (*arg == 'v') verbose++ ;
      else usage(pgm) ;
    } else {
      break ;
    }
    ++argv;
    --argc ;
  }
  if (argc == 1) {    
    size = atoi(argv[0]) ;
  }

  run(size,2) ;
  return 0 ;
}
